#!/usr/bin/env python3
# -*-coding:Utf-8 -*

"""
idr library
"""

name = "midr"
__all__ = ["__main__",
    "idr", "samic", "archimedean", "c_archimedean", "archimediean_plots",
    "log", "narrowpeak", "raw_matrix", "auxiliary"]

